#include "precomp.h"

#define DOTS 128

using namespace Tmpl8;

float rx = 0, ry = 0, rz = 0;
float elapsed1, elapsed2, elapsed3, elapsed4;
timer t;

//-----------------------------------------------------------
// Game::Sort
// Sort the dots according to depth
//-----------------------------------------------------------
void Game::Sort()
{
	for (int i = 0; i < DOTS; i++)
	{
		for (int j = 0; j < (DOTS - 1); j++)
		{
			if (m_Rotated[j].z > m_Rotated[j + 1].z)
			{
				vec3 h = m_Rotated[j];
				m_Rotated[j] = m_Rotated[j + 1];
				m_Rotated[j + 1] = h;
			}
		}
	}
}

//-----------------------------------------------------------
// Game::Init
// Initialize the application
//-----------------------------------------------------------
void Game::Init()
{	
	m_Dot = new Sprite( new Surface( "assets/dot.png" ), 4 );	
	m_Dot->SetFlags( Sprite::FLARE );
	m_Points = new vec3[DOTS];
	m_Rotated = new vec3[DOTS];
	for ( int i = 0; i < DOTS; i++ )
	{
		float dist, x, y, z;
		do
		{
			x = Rand( 2 ) - 1, y = Rand( 2 ) - 1, z = Rand( 2 ) - 1;
			dist = sqrtf( x * x + y * y + z * z );
		}
		while (dist > 1);
		m_Points[i] = vec3( x, y, z );
	}
}

//-----------------------------------------------------------
// Game::Transform
// Rotate the dots around the origin using matrix m
//-----------------------------------------------------------
void Game::Transform()
{
	mat4 m;
	m = mat4::rotatex( rx ) * mat4::rotatey( ry ) * mat4::rotatez( rz );
	rx += 0.002f, ry += 0.0035f, rz += 0.001f;
	if (rx > 360) rx -= 360;
	if (ry > 360) ry -= 360;
	if (rz > 360) rz -= 360;
	for ( int i = 0; i < DOTS; i++ ) 
	{
		m_Rotated[i] = (vec4( m_Points[i], 1 ) * m).xyz;
		// HACK: encode dot index in rotated z coordinate
		uint* hack = (uint*)&m_Rotated[i].z;
		*hack = (*hack & -4) + i; 
	}
}

//-----------------------------------------------------------
// Game::Render
// Draw the dots to the window surface
//-----------------------------------------------------------
void Game::Render()
{
	t.reset(); m_Surface->Clear( 0 ); elapsed4 = t.elapsed();
	for ( int i = 0; i < DOTS; i++ )
	{
		// extract dot index from sorted z-coordinate
		uint dotIdx = *(uint*)&m_Rotated[i].z & 3;
		// draw scaled sprite
		int sx = (int)(m_Rotated[i].x * 350.0f) + SCRWIDTH / 2;
		int sy = (int)(m_Rotated[i].y * 350.0f) + SCRHEIGHT / 2;
		int size = (int)((m_Rotated[i].z + 2) * 10.0f);
		m_Dot->SetFrame( dotIdx );
		m_Dot->DrawScaled( sx - size / 2, sy - size / 2, size, size, m_Surface );
	}
}
//-----------------------------------------------------------
// Game::Tick
// Main function of application
//-----------------------------------------------------------
void Game::Tick( float a_DT )
{
	
	t.reset(); Transform(); elapsed1 = t.elapsed();
	t.reset(); Sort(); elapsed2 = t.elapsed();
	t.reset(); Render(); elapsed3 = t.elapsed();

	char t[200];
	sprintf( t, "transform: %7.4f", elapsed1 );
	m_Surface->Print( t, 2, 2, 0x11ffffff );
	sprintf( t, "sort:      %7.4f", elapsed2 );
	m_Surface->Print( t, 2, 12, 0x11ffffff );
	sprintf( t, "render:    %7.4f", elapsed3 );
	m_Surface->Print( t, 2, 22, 0x11ffffff );
	sprintf( t, "clear:     %7.4f", elapsed4 );
	m_Surface->Print( t, 2, 32, 0x11ffffff );
}